/*
 * Copyright (c) 2016 LabKey Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.labkey.labwork.workflow;

import org.jetbrains.annotations.Nullable;
import org.labkey.api.security.Group;
import org.labkey.api.security.GroupManager;
import org.labkey.api.security.MemberType;
import org.labkey.api.security.SecurityManager;
import org.labkey.api.security.User;
import org.labkey.api.util.emailTemplate.EmailTemplate;
import org.labkey.api.util.emailTemplate.EmailTemplateService;
import org.labkey.api.workflow.NotificationConfig;
import org.labkey.api.workflow.WorkflowProcess;
import org.labkey.security.xml.GroupEnumType;

import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.labkey.labwork.workflow.WorkflowHandler.ASSAY_A_NAME;
import static org.labkey.labwork.workflow.WorkflowHandler.ASSAY_B_NAME;

/**
 * Configuration class for notifications about different stages in the process.
 */
public class WorkNotificationConfig extends NotificationConfig
{
    /**
     * Variables are provided so they may be used to determine which email configuration to use or
     * to incorporate context into the notification
     * @param variables set of process variables
     */
    public WorkNotificationConfig(Map<String, Object> variables)
    {
        super(variables);
    }

    /**
     * @return the set of users who are to be notified
     */
    @Override
    public List<User> getUsers()
    {
        Group group = GroupManager.getGroup(_container, "Scientists", GroupEnumType.PROJECT);
        if (group != null)
        {
            return SecurityManager.getAllGroupMembers(group, MemberType.ACTIVE_USERS, group.isUsers()).stream().collect(Collectors.toList());
        }

        return Collections.emptyList();
    }

    /**
     * Returns the appropriate email template based on the state of the workflow.  If task variables were used to
     * distinguish the differnt user tasks, the email template could provide further context.  N.B. Instead of
     * adding logic here to inspect this state, you could define the workflow with separate notification tasks that
     * reference separate NotificationConfig classes for place that a notification should happen.
     * @param processInstanceId identifier for the active process instance
     * @param variables the variables associated with the current execution
     * @return the template for the email to be sent
     */
    @Override
    public EmailTemplate getEmailTemplate(String processInstanceId, Map<String, Object> variables)
    {
        if (WorkflowHandler.isNewRequest(processInstanceId, _container))
        {
            WorkRequestEmailTemplate template = EmailTemplateService.get().getEmailTemplate(WorkRequestEmailTemplate.class, _container);

            template.init(
                    _initiator,
                    (Date) getVariable(WorkflowProcess.CREATED_DATE),
                    (String) getVariable(ASSAY_A_NAME),
                    (String) getVariable(ASSAY_B_NAME));
            return template;
        }
        else if (WorkflowHandler.isAssayRunActive(processInstanceId, _container))
        {
            WorkDelayedEmailTemplate template = EmailTemplateService.get().getEmailTemplate(WorkDelayedEmailTemplate.class, _container);

            template.init(
                    _initiator,
                    (Date) getVariable(WorkflowProcess.CREATED_DATE),
                    "run");
            return template;
        }
        else
            return null;
    }

    @Nullable
    @Override
    public User getLogUser()
    {
        return _initiator;
    }
}
