/*
 * Copyright (c) 2016 LabKey Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.labkey.labwork.workflow;

import org.labkey.api.data.Container;
import org.labkey.api.security.User;
import org.labkey.api.services.ServiceRegistry;
import org.labkey.api.workflow.WorkflowProcess;
import org.labkey.api.workflow.WorkflowService;
import org.labkey.api.workflow.WorkflowTask;
import org.labkey.labwork.LabWorkController;
import org.labkey.labwork.LabWorkModule;

import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Methods and constants for use in executing various labWorkflow tasks.
 */
public class WorkflowHandler
{
    public static final String PROCESS_KEY = "labWorkflow";
    public static final String ASSAY_A_KEY = "runAssayA";
    public static final String ASSAY_B_KEY = "runAssayB";
    public static final String ASSAY_A_NAME = "assayA";
    public static final String ASSAY_B_NAME = "assayB";
    public static final String ARCHIVE_METADATA = "archive";


    private static Set<String> ASSAY_RUN_KEYS = new HashSet<>();
    static
    {
        ASSAY_RUN_KEYS.add(ASSAY_A_KEY);
        ASSAY_RUN_KEYS.add(ASSAY_B_KEY);
    }

    private static final WorkflowService _service = ServiceRegistry.get(WorkflowService.class);

    /**
     * Starts a workflow with the different assay requests.
     * @param bean contains the data for initiating a request
     * @param container the container in which the request is being made
     * @param user the user making the request
     * @return the initiated process instance
     * @throws FileNotFoundException when the process definition file is not found.
     */
    public static WorkflowProcess startWorkflow(LabWorkController.WorkRequestBean bean, Container container, User user) throws FileNotFoundException
    {
        Map<String, Object> processVariables = new HashMap<>();
        processVariables.put(ASSAY_A_NAME, bean.getAssayA());
        processVariables.put(ASSAY_B_NAME, bean.getAssayB());

        processVariables.put(WorkflowProcess.INITIATOR_ID, String.valueOf(user.getUserId()));
        processVariables.put(WorkflowProcess.CONTAINER_ID, String.valueOf(container.getId()));

        String processInstanceId = _service.startWorkflow(LabWorkModule.NAME, PROCESS_KEY, "Lab work for " + user.getDisplayName(user), processVariables, container);
        return _service.getWorkflowProcess(processInstanceId);
    }

    /**
     * Determines if this processInstanceId is a new request (i.e., one that is not yet stored in the database
     * and thus has no tasks)
     * @param processInstanceId identifier for the process that is being started
     * @param container container in which the process has started
     * @return true if this is a new process instance, false otherwise
     */
    public static boolean isNewRequest(String processInstanceId, Container container)
    {
        List<WorkflowTask> tasks = _service.getCurrentProcessTasks(processInstanceId, container);
        return (tasks.size() == 0);
    }


    /**
     * Retrieves the set of tasks with a certain set of task keys that are in progress for a given process instance id
     * @param processInstanceId identifier of the process instance
     * @param container container in which the process instance is active
     * @param taskKeys the set of keys (identifiers) for the types of tasks that are of interest
     * @return a list of task names that are active with the given keys
     */
    public static List<String> getTasksInProgress(String processInstanceId, Container container, Set<String> taskKeys)
    {
        List<WorkflowTask> tasks = _service.getCurrentProcessTasks(processInstanceId, container);

        List<String> inProgress = new ArrayList<>();
        tasks.stream().filter(task -> taskKeys.contains(task.getTaskDefinitionKey())).forEach(task -> inProgress.add(task.getName()));
        return inProgress;
    }

    /**
     * Determines if one of the assay run tasks is in progress or not.  In this simple example, there is no way to
     * distinguish the two assay run tasks.  This could be done by attaching task variables to the individual tasks
     * and querying them.  Another approach would be to have separate notification tasks for each of the assay run tasks,
     * obviating the need for this method.
     * @param processInstanceId identifier for the active process instance
     * @param container container in which the process is active
     * @return true if one of the assay run tasks is in progress; false otherwise
     */
    public static boolean isAssayRunActive(String processInstanceId, Container container)
    {
        return !getTasksInProgress(processInstanceId, container, ASSAY_RUN_KEYS).isEmpty();
    }

    /**
     * This task could activate a more interesting archiving action that perhaps uses some data management capabilities.
     * In the end, you would probably want to store some primitive data types with the workflow variables to indicate,
     * say, when and where the archiving has happened.  N.B.  You generally want to store only primitive types in the
     * process variables (identifiers, dates, etc.) because if a non-primitive type is used you may have problems with
     * deserialization if the class changes.
     * @param process the process to which the archive task belongs
     * @param container the container in which the process is active.
     */
    public static void archiveResults(WorkflowProcess process, Container container)
    {
        Map<String, Object> variables = process.getVariables();
        if (variables == null)
            variables = new HashMap<>();
        variables.put(ARCHIVE_METADATA, new Date());
        _service.updateProcessVariables(process.getCurrentTasks().get(0).getId(), variables);
    }

}

